#!/usr/bin/env python3

from pathlib import Path
from xd_cwl_utils.classes.metadata.tool_metadata import ParentToolMetadata
from xd_cwl_utils.classes.cwl.make_cwl import initialize_command_line_tool_file_tool
from xd_cwl_utils.helpers.get_paths import get_tool_common_dir, main_tool_subtool_name, get_tool_metadata, get_tool_dir


def add_tool(tool_name, version_name, subtool_names=None, biotools_id=None, has_primary=False, root_repo_path=Path.cwd(), init_cwl=False):
    """
    Make the correct directory structure for adding a new command line tool. Optionally, create initialized CWL
    and metadata files. Run from cwl-tools directory.
    :param tool_name(str): Name of the tool
    :param version_name(str): version of the tool
    :param subtool_names(list(str)): list of subtool names if the tool is broken into multiple subtools.
    :param mk_meta_files(bool): Specify whether to make initial CWL and metadata files.
    :return: None
    """
    version_name = str(version_name) # In case ArgumentParser is bypassed.
    if subtool_names:
        if isinstance(subtool_names, str):
            subtool_names = [subtool_names]
    common_dir = get_tool_common_dir(tool_name, version_name, base_dir=root_repo_path)
    common_dir.mkdir(parents=True, exist_ok=False)
    if has_primary:  # Need to append __main__ onto subtools.
        if subtool_names:
            subtool_names.append(main_tool_subtool_name)
        else:
            subtool_names = [main_tool_subtool_name]
    if biotools_id:
        # tool_name will be ignored.
        parent_metadata = ParentToolMetadata.create_from_biotools(biotools_id, version_name, subtool_names, tool_name=tool_name)
    else:
        parent_metadata = ParentToolMetadata(name=tool_name, softwareVersion={'versionName': version_name, 'includedVersions': []}, featureList=subtool_names)
    if parent_metadata.featureList:
        for subtool in parent_metadata.featureList:
            subtool_obj = parent_metadata.make_subtool_metadata(subtool_name=subtool)
            subtool_obj.mk_file(base_dir=root_repo_path)
            subtool_dir = get_tool_dir(tool_name, version_name, subtool, base_dir=root_repo_path)
            instances_dir = subtool_dir / 'instances'
            instances_dir.mkdir()
            git_keep_file = instances_dir / '.gitkeep'
            git_keep_file.touch()
            if init_cwl:
                initialize_command_line_tool_file_tool(tool_name, version_name, subtool, base_dir=root_repo_path)
    parent_metadata.mk_file(root_repo_path)
    return


def add_subtool(tool_name, tool_version, subtool_name, root_repo_path=Path.cwd(), update_featureList=False, init_cwl=False):
    """
    Add subtool to already existing ToolLibrary (ParentTool file already exists)
    :param tool_name(str):
    :param version_name (str):
    :param subtool_name (str):
    :param root_repo_path (Path):
    :param update_featureList (Bool): If True, subtool does not need to be in ParentTool featureList and ParentTool will be updated. Will throw error if False and subtool is not in ParentTool featureList.
    :param init_cwl:
    :return:
    """
    parent_path = get_tool_metadata(tool_name, tool_version, parent=True, base_dir=root_repo_path)
    parent_meta = ParentToolMetadata.load_from_file(parent_path)
    if update_featureList:
        if parent_meta.featureList is None:
            parent_meta.featureList = [subtool_name]
        else:
            if not subtool_name in parent_meta.featureList:
                parent_meta.featureList.append(subtool_name)
        parent_meta.mk_file(base_dir=root_repo_path)  # Remake the file. Needs to be remade if updated.

    subtool_meta = parent_meta.make_subtool_metadata(subtool_name)
    subtool_meta.mk_file(base_dir=root_repo_path)
    subtool_dir = get_tool_dir(tool_name, tool_version, subtool_name, base_dir=root_repo_path)
    instances_dir = subtool_dir / 'instances'
    instances_dir.mkdir()
    git_keep_file = instances_dir / '.gitkeep'
    git_keep_file.touch()
    if init_cwl:
        initialize_command_line_tool_file_tool(tool_name, tool_version, subtool_name, base_dir=root_repo_path)
    return



#!/usr/bin/env python3

import argparse
import sys
import logging
from pathlib import Path
from xd_cwl_utils.validate import validate_parent_tool_metadata, validate_subtool_metadata, validate_script_metadata, \
    validate_common_script_metadata, validate_workflow_metadata, validate_tools_dir, validate_scripts_dir, validate_workflows_dir, validate_repo


def get_parser():
    parser = argparse.ArgumentParser(description="Validate metadata files.")
    subparsers = parser.add_subparsers(description="Specify type of metadata to validate.", dest='command')
    parser.add_argument('-p', '--root_repo_path', dest='root_path', type=Path, default=Path.cwd(),
                        help="Specify the root path of your cwl content repo if it is not the current working directory.")

    validate_parent_tool = subparsers.add_parser('parent_tool', help="Validate parent tool metadata.")
    validate_parent_tool.add_argument('path', type=Path, help="Path to parent tool metadata file.")

    validate_subtool = subparsers.add_parser('subtool', help="Validate subtool metadata.")
    validate_subtool.add_argument('path', type=Path, help="Path to subtool metadata file.")

    validate_script = subparsers.add_parser('script', help="Validate script metadata")
    validate_script.add_argument('path', type=Path, help="Path to script metadata file.")

    validate_script_common = subparsers.add_parser('script_common', help="Validate common script metadata.")
    validate_script_common.add_argument('path', type=Path, help="Validate common script metadata")

    validate_workflow = subparsers.add_parser('workflow', help="Validate workflow metadata.")
    validate_workflow.add_argument('path', type=Path, help="Path to workflow metadata")

    validate_tool_dir = subparsers.add_parser('tool_dir', help='Validate an entire cwl-tools directory.')

    validate_script_dir = subparsers.add_parser('script_dir', help='Validate and entire cwl-scripts directory')

    validate_workflow_dir = subparsers.add_parser('workflow_dir', help="validate and entire cwl-workflows directory")

    validate_repo = subparsers.add_parser('repo', help='Validate entire cwl source repo.')

    return parser


def main(argsl=None):
    if not argsl:
        argsl = sys.argv[1:]

    parser = get_parser()
    args = parser.parse_args(argsl)
    command = args.command
    try:
        if args.path.is_absolute():
            full_path = args.path
        else:
            full_path = args.root_path / args.path
    except AttributeError as e:
        if not command in ('tool_dir', 'script_dir', 'workflow_dir', 'repo'):
            raise
    if command == 'parent_tool':
        validate_parent_tool_metadata(full_path)
    elif command == 'subtool':
        validate_subtool_metadata(args.path)
    elif command == 'script':
        try:
            validate_script_metadata(args.path)
        except:
            logging.error(f"Problem with {args.path}")
            raise
    elif command == 'script_common':
        validate_common_script_metadata(args.path)
    elif command == 'workflow':
        validate_workflow_metadata(args.path)

    elif command == 'tool_dir':
        validate_tools_dir(base_dir=args.root_path)
    elif command == 'script_dir':
        validate_scripts_dir(base_dir=args.root_path)
    elif command == 'workflow_dir':
        validate_workflows_dir(base_dir=args.root_path)
    elif command == 'repo':
        validate_repo(base_dir=args.root_path)
    else:
        parser.print_help()
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))




import logging
import cwltool.main
from ..classes.cwl.command_line_tool import load_document

def validate_cwl_tool(cwl_doc_path):
    # Just load into CommandLineTool class and see if it works. Seems to be good.
    cwl_doc_path = str(cwl_doc_path)
    command_line_tool = load_document(cwl_doc_path)
    return



import tempfile
import logging
from pathlib import Path
from ruamel.yaml import safe_load
from semantic_version import Version
from xd_cwl_utils.classes.metadata.tool_metadata import ParentToolMetadata, SubtoolMetadata
from xd_cwl_utils.classes.metadata.script_metadata import ScriptMetadata, CommonScriptMetadata
from xd_cwl_utils.classes.metadata.workflow_metadata import WorkflowMetadata
from .content_maps import make_tools_map, make_script_maps, make_workflow_maps
from .helpers.get_paths import get_metadata_path
from .helpers.validate_cwl import validate_cwl_tool
from .validate_inputs import validate_all_inputs_for_tool
from .classes.cwl.command_line_tool import ValidationException

def metadata_validator_factory(class_to_validate):
    def metadata_validator(metadata_path):
        try:
            metadata_instance = class_to_validate.load_from_file(metadata_path)
            # print(f"Metadata in {metadata_path} is valid {str(class_to_validate)}")
            logging.info(f"Metadata in {metadata_path} is valid {str(class_to_validate)}")
        except:
            logging.error(f"{str(class_to_validate)} in {metadata_path} failed validation.")
            raise
        return
    return metadata_validator

validate_parent_tool_metadata = metadata_validator_factory(ParentToolMetadata)
validate_subtool_metadata = metadata_validator_factory(SubtoolMetadata)
validate_script_metadata = metadata_validator_factory(ScriptMetadata)
validate_common_script_metadata = metadata_validator_factory(CommonScriptMetadata)
validate_workflow_metadata = metadata_validator_factory(WorkflowMetadata)

def validate_tools_dir(base_dir=None):
    """
    Validate all cwl files, metadata files, instances and instance metadata in a cwl-tools directory
    :return:
    """
    tool_map_temp_file = tempfile.NamedTemporaryFile(prefix='tools_map', suffix='.yaml', delete=True)  # Change to False if file doesn't persist long enough.
    make_tools_map(tool_map_temp_file.name, base_dir=base_dir)
    with tool_map_temp_file as tool_map:
        tool_map_dict = safe_load(tool_map)
    for identifier, values in tool_map_dict.items():
        tool_path = base_dir / values['path']
        tool_type = values['type']
        metadata_status = values['metadataStatus']


        if tool_type == 'parent':
            if not 'common' in tool_path.parts:
                raise ValueError  # an extra check.
            meta_type = 'parent_tool'  # correspond to command in validate_metadata
            validate_parent_tool_metadata(tool_path)
            # assert no instances directory here?
        else:  # is a subtool.
            assert tool_type == 'subtool'
            cwl_status = values['cwlStatus']
            # validate metadata
            metadata_path = get_metadata_path(tool_path)
            validate_subtool_metadata(metadata_path)

            # validate cwl only if metadata specifies if it is version Released
            if cwl_status in ('Draft', 'Released'):
                validate_cwl_tool(tool_path)

                # validate instances
                validate_all_inputs_for_tool(tool_path)

    return


def validate_scripts_dir(base_dir=None):

    script_map_temp_file = tempfile.NamedTemporaryFile(prefix='scripts_map', suffix='.yaml', delete=True)  # Change to False if file doesn't persist long enough.
    make_script_maps(script_map_temp_file.name, base_dir=base_dir)
    with script_map_temp_file as script_map:
        script_map_dict = safe_load(script_map)
    for identifier, values in script_map_dict.items():
        # validate metadata
        script_path = base_dir / values['path']
        metadata_path = get_metadata_path(script_path)
        validate_script_metadata(metadata_path)

        # validate cwl
        cwl_status = values['cwlStatus']
        if cwl_status in ('Draft', 'Released'):
            validate_cwl_tool(script_path)
            validate_all_inputs_for_tool(script_path)
    return


def validate_workflows_dir(base_dir=None):
    workflow_map_temp_file = tempfile.NamedTemporaryFile(prefix='workflows_map', suffix='.yaml', delete=True)
    make_workflow_maps(workflow_map_temp_file.name, base_dir=base_dir)
    with workflow_map_temp_file as workflow_map:
        workflow_map_dict = safe_load(workflow_map)
    for identifier, values in workflow_map_dict.items():
        workflow_path = base_dir / values['path']
        workflow_metadata = get_metadata_path(workflow_path)
        validate_workflow_metadata(workflow_metadata)

        cwl_status = values['cwlStatus']
        if cwl_status in ('Draft', 'Released'):
            print(f"Make sure you validate {workflow_path}")  # Todo. Think I have good way to validate somewhere in biodrafter. Need to port here (needs to be put in a temporary directory with the tools and workflows that it calls.)
    return

def validate_repo(base_dir=None):
    validate_tools_dir(base_dir=base_dir)
    validate_scripts_dir(base_dir=base_dir)
    validate_workflows_dir(base_dir=base_dir)
    return



import tempfile
from unittest import TestCase
import os
from xd_cwl_utils.config import config
from xd_cwl_utils.content_maps import make_tools_map, make_workflow_maps, make_script_maps

test_constants = {'script_group1': 'ENCODE-DCC', 'script_version1': '1.1.x', 'script_project1': 'atac-seq-pipeline',
                  'test_software_version': {'versionName': 'test_version'}}


class TestBase(TestCase):

    def get_content_map_paths(self):
        return {'tool_maps': config[os.environ.get('CONFIG_KEY')]['content_maps_dir'] / 'tool-maps.yaml',
                'script_maps': config[os.environ.get('CONFIG_KEY')]['content_maps_dir'] / 'script-maps.yaml',
                'workflow_maps': config[os.environ.get('CONFIG_KEY')]['content_maps_dir'] / 'workflow-maps.yaml'}

    def update_tool_maps(self):
        outfile_path = self.get_content_map_paths()['tool_maps']
        make_tools_map(outfile_path)

    def update_script_maps(self):
        outfile_path = self.get_content_map_paths()['script_maps']
        make_script_maps(outfile_path)

    def update_workflow_maps(self):
        outfile_path = self.get_content_map_paths()['workflow_maps']
        make_workflow_maps(outfile_path)

    def setUp(self):
        self.test_dir = tempfile.TemporaryDirectory()

    def tearDown(self) -> None:
        pass





# from cwlgen import CommandLineTool, CommandInputParameter, CommandOutputParameter, CommandOutputBinding
from ruamel.yaml import YAML, tokens, error
from ruamel.yaml.comments import CommentedMap
from xd_cwl_utils.helpers.get_paths import get_cwl_tool, get_cwl_script, main_tool_subtool_name

blank_line_tk = tokens.CommentToken('\n\n', error.CommentMark(0), None)

# def initialize_commmand_line_tool_file_cwl_gen(tool_name, version_name, subtool_name, base_dir):
#     cwl_tool_path = get_cwl_tool(tool_name, version_name, subtool_name=subtool_name, base_dir=base_dir)
#     base_command = f"{tool_name} {subtool_name}" if subtool_name != main_tool_subtool_name else tool_name
#     clt_obj = CommandLineTool(base_command=base_command, cwl_version="v1.0")
#     clt_obj.inputs.append(CommandInputParameter('input1', param_type='string'))
#     clt_obj.outputs.append(CommandOutputParameter('output1', param_type='File', doc="Add documentation here", output_binding=CommandOutputBinding()))
#     clt_obj.export(str(cwl_tool_path))



def _initialize_command_line_tool_file_yaml(base_command, cwl_path):
    cwl_map = CommentedMap()
    doc_placholder = 'Document here.'
    input_name = 'input1'
    output_name = 'output1'
    cwl_map['cwlVersion'] = 'v1.0'
    cwl_map['class'] = 'CommandLineTool'
    cwl_map['baseCommand'] = base_command
    cwl_map['stdout'] = None
    cwl_map['inputs'] = CommentedMap([(input_name, {'type': None, 'inputBinding': {'position': 1, 'prefix': None}, 'doc': doc_placholder})])
    cwl_map['outputs'] = CommentedMap([(output_name, {'type': None, 'outputBinding': {'glob': None}, 'doc': doc_placholder})])
    # insert some blank lines.
    cwl_map.ca.items['stdout'] = [None, None, blank_line_tk, None]
    cwl_map.ca.items['outputs'] = [None, [blank_line_tk], None, None]

    yaml = YAML(pure=True)
    yaml.default_flow_style = False
    yaml.indent(mapping=2, sequence=4, offset=2)
    with cwl_path.open('w') as cwl_file:
        yaml.dump(cwl_map, cwl_file)
    return

def initialize_command_line_tool_file_tool(tool_name, version_name, subtool_name, base_dir):
    cwl_tool_path = get_cwl_tool(tool_name, version_name, subtool_name=subtool_name, base_dir=base_dir)
    base_command = f"{tool_name} {subtool_name}" if subtool_name != main_tool_subtool_name else tool_name
    _initialize_command_line_tool_file_yaml(base_command, cwl_tool_path)
    return

def initialize_command_line_tool_file_script(group_name, project_name, script_version, script_name, base_dir):
    cwl_script_path = get_cwl_script(group_name, project_name, script_version, script_name, base_dir)
    base_command = script_name
    _initialize_command_line_tool_file_yaml(base_command, cwl_script_path)
    return



from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import skip
from tests.test_base import TestBase
from xd_cwl_utils.add_content import main as add_content_main

# @skip('')
class TestAddToolMain(TestBase):
    # @skip('')
    def test_add_tool(self):
        with TemporaryDirectory() as tmp_dir:
            tool_name = 'test_1'
            tool_version = 'fake.1'
            temp_path = Path(tmp_dir)
            new_directory = temp_path / 'cwl-source'
            new_directory.mkdir()
            add_content_main([f"-p {new_directory}", 'tool', tool_name, tool_version, f"--has_primary"])
        return

    # @skip('')
    def test_add_tool_with_subtools(self):
        with TemporaryDirectory() as tmp_dir:
            tool_name = 'test_2'
            tool_version = 'fake.2'
            subtools = ['subtool1', 'subtool2', 'subtool3']
            temp_path = Path(tmp_dir)
            new_directory = temp_path / 'cwl-source'
            new_directory.mkdir()
            add_content_main([f"-p {new_directory}", 'tool', tool_name, tool_version] + subtools)
        return

    # @skip('')
    def test_add_tool_with_biotools_id(self):
        with TemporaryDirectory() as tmp_dir:
            tool_name = 'test_biotools'
            tool_version = 'fake.3'
            biotools_id = 'malvirus'
            subtools = ['subtool1', 'subtool2', 'subtool3']
            options = ['--biotoolsID', biotools_id]
            temp_path = Path(tmp_dir)
            new_directory = temp_path / 'cwl-source'
            new_directory.mkdir()
            add_content_main([f"-p {new_directory}", 'tool', tool_name, tool_version] + subtools + options)
        return

    def test_add_subtool(self):
        with TemporaryDirectory() as tmp_dir:
            tool_name = 'test_4'
            tool_version = 'fake.4'
            subtool_name = 'new_subtool'
            temp_path = Path(tmp_dir)
            new_directory = temp_path / 'cwl-source'
            new_directory.mkdir()
            add_content_main([f"-p {new_directory}", 'tool', tool_name, tool_version])
            add_content_main([f"-p {new_directory}", 'subtool', tool_name, tool_version, subtool_name, '-u'])
        return


class TestAddScriptMain(TestBase):

    def test_add_common_script(self):
        with TemporaryDirectory() as tmp_dir:
            group_name = 'test_group1'
            project_name = 'fake_project_1'
            script_version = '1.nope'
            file_name = "some_filename"
            temp_path = Path(tmp_dir)
            new_directory = temp_path / 'cwl-scripts'
            new_directory.mkdir()
            add_content_main([f"-p {new_directory}", 'common_script', group_name, project_name, script_version, file_name])
        return

    def test_add_script(self):
        with TemporaryDirectory() as tmp_dir:
            group_name = 'test_group2'
            project_name = 'fake_project_2'
            script_name = 'new_script_2'
            script_version = '2.nope'
            temp_path = Path(tmp_dir)
            new_directory = temp_path / 'cwl-scripts'
            new_directory.mkdir()
            add_content_main([f"-p {new_directory}", 'script', group_name, project_name, script_version, script_name])
        return




from tempfile import TemporaryDirectory
from tests.test_base import TestBase
from xd_cwl_utils.add.add_tools import add_tool, add_subtool

class TestAddTool(TestBase):

    def test_add_parent_only(self):
        with TemporaryDirectory() as tmp_dir:
            tool_name = 'test__1'
            tool_version = '123'
            new_path = add_tool(tool_name, tool_version, root_repo_path=tmp_dir)
        # print(new_path)
        return



    def test_add_with_main(self):
        with TemporaryDirectory() as tmp_dir:
            tool_name = 'test___2'
            tool_version = '123'
            new_path = add_tool(tool_name, tool_version, has_primary=True, root_repo_path=tmp_dir)
            # print(new_path)
        return


    def test_add_with_subtools(self):
        with TemporaryDirectory() as tmp_dir:
            tool_name = 'test___3'
            tool_version = '123'
            new_path = add_tool(tool_name, tool_version, subtool_names=['first', 'second'], has_primary=True, root_repo_path=tmp_dir)
            # print(new_path)
        return

    def test_add_parent_then_subtool(self):
        with TemporaryDirectory() as tmp_dir:
            tool_name = 'test_tool'
            tool_version = '0.1'
            add_tool(tool_name, tool_version, root_repo_path=tmp_dir)
            add_subtool(tool_name, tool_version, 'subtool1', root_repo_path=tmp_dir, update_featureList=True)
            assert True
        return


    def test_add_multiple(self):
        with TemporaryDirectory() as tmp_dir:
            add_tool('fake_1', '123', root_repo_path=tmp_dir)  # parent only.
            add_tool('fake_2', '123', has_primary=True, root_repo_path=tmp_dir)  # # parent with main tool
            add_tool('fake_3', '123', subtool_names=['first', 'second'], has_primary=True,
                     root_repo_path=tmp_dir)
            assert True
        return